/* Request.pmod
 *
 */

#include "Types.h"

#define WERROR werror

import Types;

mapping (int:program) requests;

void register(int command, program p)
{
  requests[command] = p;
}

program query_program(int n)
{
  return requests[n];
}

class Request {
  import Types;

  inherit Generic;
  import protocol_A;
  
  int ref_no;
  int command;
  object session;
  object connection;
  array(object) tokens;
  
  void create(int n, int cmd)
  {
    ref_no = n;
    command = cmd;
    WERROR(sprintf("Request->create: ref_no = %d, cmd = %d\n", ref_no, cmd));
  }

  void set_context(object s, object c)
  {
    session = s;
    connection = c;
  }

  object process_tokens()
  {
    // trace(1);
    object o = init_protA(tokens);
    // trace(0);
    return o;
  }
  
  void set_tokens(array(object) t)
  {
    tokens = t;
  }

  void answer_low(array(string) data)
  {
    connection->answer(answer_to_A(ref_no, data));
  }
  
  void answer(object ...args)
  {
    answer_low(Array.map(args, any_to_A));
  }
    
  void error(int error_no, int error_status)
  {
    connection->answer(error_to_A(ref_no, error_no, error_status));
  }

  void protocol_error()
  {
    connection->answer(protocol_error_to_A());
  }

  void process() {}

  void process_start()
  {
    if (!process_tokens())
      {
	protocol_error();
	return;
      }
    process();
  }
}

class Get_Time
{
  inherit Request;

  void process()
  {
    answer_low( ({ protocol_A.time_to_A(time()) }) );
  }
}
    
      
constant login_is_invisible = 1;

class Login
{
  inherit Request;
  constant fields = ({ FIELD("int", "person"),
		       FIELD("string", "passwd"),
		       FIELD3("bitstring", "visibility", 1) });

  constant need_sync_before = 1;
  constant need_sync_after = 1;

  int person;
  string passwd;
  bitstring_t visibility;
  
  void process()
  {
    int err = session->login(person, passwd,
			     visibility & login_is_invisible);
    if (err)
      error(err, 0);
    else
      answer();
  }
}

class Login_Old
{
  inherit Request;
  constant fields = ({ FIELD("int", "person"),
		       FIELD("string", "passwd") });

  constant need_sync_before = 1;
  constant need_sync_after = 1;

  int person;
  string passwd;

  void process()
  {
    int err = session->login(person, passwd, 0);
    if (err)
      error(err, 0);
    else
      answer();
  }
}

class Change_What_Am_I_Doing
{
  inherit Request;
  constant fields = ({ FIELD("string", "msg") });
  string msg;

  void process()
  {
    int err = session->change_what_am_i_doing(msg);

    if (err)
      error(err, 0);
    else
      answer();
  }
}

class User_Active
{
  inherit Request;

  void process()
  {
    session->user_active();
    answer();
  }
}

class Who_Is_On_Dynamic
{
  inherit Request;
  constant fields = ({ FIELD("int", "want_visible"),
                       FIELD("int", "want_invisible"),
                       FIELD("int", "active_last") });
  int want_visible;
  int want_invisible;
  int active_last;

  void process()
  {
    answer_low(({ protocol_A.array_to_A_low(session
					    ->who_is_on_dynamic(want_visible,
								want_invisible,
								active_last))
		}));
  }
}

class Get_Static_Session_Info
{
  inherit Request;
  constant fields = ({ FIELD("int", "session_no") });
  int session_no;

  void process()
  {
    string s = session->get_static_session_info(session_no);
    //    WERROR("Get_Static_Session_Info: s = " + s + "\n");
    if (!s)
      error(Errors.undef_session , 0);
    answer_low(({ s }));
  }
}

class Old_Create_Person
{
  inherit Request;
  constant fields = ({ FIELD("string", "name"),
		       FIELD("string", "passwd") });
  
  string name;
  string passwd;

  void process()
  {
    array res = session->create_person(name, passwd);

    if (res[0])
      error(res[0], 0);
    else
      answer(res[1]);
  }
}

class Old_Create_Conf
{
  inherit Request;
  constant fields = ({ FIELD("string", "name"),
		       FIELD3("bitstring", "type", 8) });

  string name;
  int type;

  void process()
  {
    array res = session->create_conf(name, type);

    if (res[0])
      error(res[0], 0);
    else
      answer(res[1]);
  }
}

class Old_Create_Text
{
  inherit Request;
  constant fields = ({ FIELD("string", "text"),
		       FIELD3("array", "misc_info", Server.Misc_Info) });
  string text;
  array misc_info;

  void process()
  {
    string subject;
    
    int i = search(text, "\n");
    if (i<0)
      {
	subject = text;
	text = "";
      }
    else
      {
	subject = text[..i-1];
	text = text[i+1..];
      }
    
    array res = session->create_text(subject, text, misc_info);

    if (res[0])
      error(@res);
    else
      answer(res[1]);
  }
}

class Get_Text
{
  inherit Request;
  constant fields = ({ FIELD("int", "text_no"),
		       FIELD("int", "start_char"),
		       FIELD("int", "end_char") });

  int text_no;
  int start_char;
  int end_char;
  
  void process()
  {
    array res = session->get_text_interval(text_no, start_char, end_char);
    if (res[0])
      error(res[0], 0);
    else
      answer(res[1]);
  }
}

class Old_Get_Text_Stat
{
  inherit Request;
  constant fields = ({ FIELD("int", "text_no") });

  int text_no;

  void process()
  {
    array res = session->get_text_stat(text_no);
    if (res[0])
      error(res[0], 0);
    else
      answer(res[1]);
  }
}
    
class Add_Member
{
  inherit Request;
  constant fields = ({ FIELD("int", "conf_no"),
		       FIELD("int", "pers_no"),
		       FIELD("int", "priority"),
		       FIELD("int", "where") });

  constant need_sync_after = 1;

  int conf_no;
  int pers_no;
  int priority;
  int where;
  
  void process()
  {
    int res = session->add_member(conf_no, pers_no, priority, where);
    if (res)
      error(res, 0);
    else
      answer();
  }
}
		       
class Get_Membership
{
  inherit Request;
  constant fields = ({ FIELD("int", "person"),
		       FIELD("int", "first"),
		       FIELD("int", "no_of_confs"),
		       FIELD3("bitstring", "mask", 1) });
  
  int person;
  int first;
  int no_of_confs;
  int mask;

  void process()
  {
    array res = session->get_memberships(person, first, no_of_confs, mask);

    if (res[0])
      error(res[0], 0);
    else
      answer(res[1]);
  }
}

class Query_Read_Texts
{
  inherit Request;
  constant fields = ({ FIELD("int", "person"),
		       FIELD("int", "conference") });
  int person;
  int conference;

  void process()
  {
    array res = session->query_membership(person, conference);

    if (res[0])
      error(res[0], 0);
    else
      answer(res[1]);
  }
}

class Get_Map
{
  inherit Request;
  constant fields = ({ FIELD("int", "conf_no"),
		       FIELD("int", "first_local_text_no"),
		       FIELD("int", "no_of_texts") });

  int conf_no;
  int first_local_text_no;
  int no_of_texts;
  
  void process()
  {
    array res = session->get_map(conf_no, first_local_text_no, no_of_texts);

    if (res[0])
      error(res[0], 0);
    else
      answer(@res[1..]);
  }
}

class Mark_As_Read
{
  inherit Request;
  constant fields = ({ FIELD("int", "conference"),
		       FIELD3("array", "texts", "int") });
  int conference;
  array(int) texts;

  void process()
  {
    int res = session->mark_as_read(texts);
    if (res)
      error(res, 0);
    else
      answer();
  }
}

class Get_Conf_Stat_Old
{
  inherit Request;
  constant fields = ({ FIELD("int", "conf_no") });

  int conf_no;
  
  void process()
  {
    array res = session->get_conf_stat(conf_no);
    if (res[0])
      error(res[0], 0);
    else
      answer(res[1]);
  }
}

class Who_Am_I
{
  inherit Request;

  void process()
  {
    answer(session->who_am_i());
  }
}

class Get_Person_Stat
{
  inherit Request;
  constant fields = ({ FIELD("int", "pers_no") });

  int pers_no;

  void process()
  {
    /* res == ({ felkod, objekt, nyckel }) */
    array res = session->get_person_stat(pers_no);
    if (res[0])
      error(res[0], 0);
    else
      {
	string response = res[1]->query_protA();
	res = 0;  /* Slpp lset */
	answer_low( ({ response }) );
      }
  }
}

class Lookup_Z_Name
{
  inherit Request;
  constant fields = ({ FIELD("string", "name"),
		       FIELD("int", "want_confs"),
		       FIELD("int", "want_pers") });
  
  string name;
  int want_confs;
  int want_pers;

  void process()
  {
    answer(session->lookup_z_name(name, want_confs, want_pers));
  }
}

class Get_Info_old
{
  inherit Request;
  
  void process()
  {
    answer(@ session->info_old());
  }
}

class Version_Info
{
  inherit Request;
  
  void process()
  {
    answer(@ session->version_info());
  }
}

/* Returnerar 0 om anropet r helt fel,
 * annars en array ({ 1,  requestobjekt }) eller
 * ({ 0, ref_no, Errors.not_impl, error_status }). */
array make_request(array(object) tokens)
{
  if ( (sizeof(tokens) < 2)
       || (tokens[0]->token != "int")
       || (tokens[1]->token != "int"))
    return 0;
  int ref_no = (int) tokens[0]->value;
  int command = (int) tokens[1]->value;

  program p = query_program(command);
  if (!p)
    return ({ 0, ref_no, Errors.not_impl, 0});
  object req = p(ref_no, command);
  req->set_tokens(tokens[2..]);
  return ({ 1, req });
}

void create()
{
  requests = ([ ]);
  register(0, Login_Old);
  // register(2, Change_Conf);
  register(4, Change_What_Am_I_Doing);
  register(5, Old_Create_Person);
  register(9, Query_Read_Texts);
  register(10, Old_Create_Conf);
  // register(12, Lookup_Name);  /* Obsolet */
  register(14, Add_Member);
  // register(23, Get_Marks);
  register(25, Get_Text);
  register(26, Old_Get_Text_Stat);
  register(27, Mark_As_Read);
  register(28, Old_Create_Text);
  register(34, Get_Map);
  register(35, Get_Time);
  register(36, Get_Info_old);
  register(46, Get_Membership);
  // register(48, Get_Members);
  register(49, Get_Person_Stat);
  register(50, Get_Conf_Stat_Old);
  register(56, Who_Am_I);
  // register(61, Find_Previous_Text_No);
  register(62, Login);
  // register(69, Set_Client_Version);
  register(75, Version_Info);
  register(76, Lookup_Z_Name);
  // register(78, Get_UConf_Stat);  /* Nr anvnds den? */
  register(82, User_Active);
  register(83, Who_Is_On_Dynamic);
  register(84, Get_Static_Session_Info);
}
