/* Server.pmod
 *
 * Objects used in the server.
 */

#include "Types.h"

#if 0
#define WERROR werror
#else
#define WERROR(x)
#endif

class DB_Object
{
  constant db_fields = ({ });
  constant db_version = 1;

  array fancy_to_db()
  {
    return ({ });
  }

  object fancy_from_db()
  {
    return this_object();
  }
  
  mixed to_db()
  {
    return rows(this_object(), db_fields) + fancy_to_db();
  }

  object from_db(mixed data)
  {
    // WERROR(sprintf("DB_Object->from_db: data = %O,\n"
    // 		      "               db_fields = %O.\n",
    // 		      data, db_fields));
    if (sizeof(data) < (sizeof(db_fields)))
      return 0;
    int i = 0;
    string name;
    foreach(db_fields, name)
      this_object()[name] = data[i++];

    return fancy_from_db(data[i..]);
  }
}

#define CHECK_LISTS

class Text_List
{
  inherit DB_Object;

  array(int) texts = ({ });

  void check_state()
  {
    for(int i = 0; i+1 < sizeof(texts); i++)
      if (texts[i] >= texts[i+1])
	throw( ({ "Server.Read_Unread_List->check_state: "
		    "List is not strictly increasing!\n",
		    backtrace() }) );
  }

  /* Delta-koda listan, fr att den ska packa bttre. */
  array to_db()
  {
#ifdef CHECK_LISTS
    check_state();
#endif
    return ({ Misc.pack_array(texts) });
  }

  object from_db(array data)
  {
    texts = Misc.unpack_array(data[0]);
#ifdef CHECK_LISTS
    check_state();
#endif
    return this_object();
  }
}

class UConference
{
  inherit DB_Object;
  inherit Types.UConference;

  /* XXX highest_local_no anger antal texter som ngonsin har funnits
   * i mtet. r det vettigt??? */
  constant db_fields = ({ "name", "type", "highest_local_no", "nice" });
}

class Conf_texts
{
  inherit DB_Object;

  inherit Text_List;
}

class Conference
{
  inherit DB_Object : db;
  constant db_fields = ({ "creation_time", "last_written",
			  "creator", "supervisor",
			  "super_conf", "presentation",
			  "permitted_submitters", "msg_of_day",
			  "no_of_texts", "members" });
  constant db_version = 2;
  
  time_t creation_time;
  time_t last_written;
  int creator;
  int presentation;
  int supervisor;
  int permitted_submitters;
  int super_conf;
  int msg_of_day;
  /* Antal texter som ngonsin har funnits i mtet */
  int no_of_texts;
  multiset(int) members = (< >);

  /* Baktkompatibilitet */
  object texts;    /* Existerande texter i mtet */

  object from_db_version(int version, array data)
  {
    WERROR(sprintf("Server.Conference->from_db_version %d.\n",
		   version));
    switch(version)
      {
      case 1:
	return (texts = Conf_texts()->from_db(data[-1]))
	  && from_db(data[..sizeof(data)-2]);
      case 2:
	return from_db(data);
      default:
	return 0;
      }
  }

  array to_db()
  {
    if (texts)
      throw( ({ "Server.Conference->to_db: "
		  "Attempt to store data using obsolete format.\n",
		  backtrace() }) );
    return db::to_db();
  }
  
#if 0
  array fancy_to_db()
  {
    return ({ Misc.pack_array(texts) });
  }

  object fancy_from_db(array data)
  {
    texts = Misc.unpack_array(data[0]);
    return this_object();
  }
#endif
}

class Conf_Z_Info
{
  inherit Types.Conf_Z_Info;
  /* Hr r nog rtt plats att cacha kanoniserade namn och snt. */
}

class Membership
{
  inherit DB_Object;
  constant db_fields = ({ "last_time_read", "conference",
			  "priority", "first_unread",
			  "no_of_texts_read", "change_count" });
  constant db_version = 2;
  
  time_t last_time_read;
  int conference;
  int priority;
  int first_unread;
  int no_of_texts_read; /* Skillnaden mellan detta tal och mtets
			 * no_of_texts ger antalet olsta. */
  int change_count;

  object from_db_version(int version, array data)
  {
    WERROR(sprintf("Server.Memberships->from_db_version %d.\n",
		   version));
    switch(version)
      {
      case 1: /* En ny variabel first_unread_change_count */
	return arrayp(data) && from_db(data + ({ 0 }) );
      case 2:
	return from_db(data);
      default:
	return 0;
      }
  }
  
  object query_old_membership()
  {
    object m = Types.Membership();
    m->last_time_read = last_time_read;
    m->conference = conference;
    m->priority = priority;
    m->last_text_read = first_unread && (first_unread - 1);
    m->read_texts = ({ });
    return m;
  }
}

#define CHECK check_state()

class Read_Unread_List
{
  import Misc;
  inherit Text_List : l;
  /* Frsta elementet r den frsta lsmarkerade texten, Nsta
   * r det frsta olsta efter det, s nsta lsmarkerade, osv.
   *
   * Allmnt r element texts[2*n] frsta lsmarkerade efter texts[2*n-1],
   * och texts[2*n+1] r frsta olsta efter texts[2*n].
   */

  /* l ska alltid innehlla ett jmnt antal tal, annars kommer ju alla
     nya texter att rknas som lsta. */
  
  constant db_version = 1;

  void check_state()
  {
    if (sizeof(l::texts) % 1)
      throw( ({ "Server.Read_Unread_List->check_state: Odd size!\n",
		  backtrace() }) );
    l::check_state();
  }
        
  int is_read(int n)
  {
#ifdef CHECK_LISTS
    check_state();
#endif
    return ((binary_search(n, l::texts) % 2) == 0);
  }

  /* Anvnds fr att bygga listorna som returneras frn get_membership. */
  array(int) query_read(array(int) texts_in_conf)
  {
#ifdef CHECK_LISTS
    check_state();
#endif
    WERROR("Server.Read_Unread_List->query_read\n");
    WERROR(sprintf("texts_in_conf = %O\n", texts_in_conf));
    WERROR(sprintf("l::texts = %O\n", l::texts));
    int i = 0;
    int j = 0;
    /* Hela tiden ska l::texts[j] <= texts_in_conf[i] */

    if (!sizeof(l::texts) || !sizeof(texts_in_conf))
      /* Inga lsta */
      return ({ });
    if (texts_in_conf[0] < l::texts[0])
      { /* Frsta texten r olst */
	i = binary_search(l::texts[0]-1, texts_in_conf)+1;
	if (i < 0)
	  throw( ({ "Server.Read_Unread_List->query_read: Internal error.\n",
		      backtrace() }) );
      }	
    while( (i < sizeof(texts_in_conf)) && (j < sizeof(l::texts) ) )
      { /* texts_in_conf[i] r nsta text av oknd status.
	 * [l::texts[j], l::texts[j+1]) r frsta intervallet som
	 * text[i] kan ligga i. Dvs l::texts[j] <= texts_in_conf[i].
	 */
	WERROR(sprintf("i = %d, j = %d\n", i, j));
	int j = binary_search(texts_in_conf[i], l::texts, j);
	/* Nu r l::texts[j] <= texts_in_conf[i] < l::texts[j+1] */
	if (j < 0)
	  throw( ({ "Server.Read_Unread_List->query_read: Internal error.\n",
		      backtrace() }) );
	if (j%2)
	  { /* texts_in_conf[i] r olst. Ta bort den, och resten av texterna
	     * i samma intervall frn listan. */
	    if ( (j == sizeof(l::texts) - 1)
		 || (i == sizeof(texts_in_conf) - 1))
	      { /* Det finns inga fler lsta. */
		texts_in_conf = texts_in_conf[..i-1];
		break;
	      }
	    /* texts_in_conf[i] < l::texts[j+1] ==> texts_in_conf[i]
               <= l::texts[j+1] - 1 */
	    int k = binary_search(l::texts[j+1]-1, texts_in_conf, i);
	    if (k < 0)
	      throw( ({ "Server.Read_Unread_List->query_read: "
			  "Internal error.\n", backtrace() }) );
	    /* Nu r texts_in_conf[k] < l::texts[j+1], dvs intervallet
	     * texts_in_conf[i..k] ska bort. */
	    texts_in_conf = texts_in_conf[..i-1] + texts_in_conf[k+1..];
	    j++;
	    /* Nu ska texts_in_conf[i] >= l::texts[j] igen. */
	  }
	else
	  { /* texts_in_conf[i] r lst. */
	    if ( (j == sizeof(l::texts) - 1)
		 || (i == sizeof(texts_in_conf) - 1))
	      /* Behll resten av listan */
	      break;
	    int k = binary_search(l::texts[j+1]-1, texts_in_conf, i) + 1;
	    if (k < 0)
	      throw( ({ "Server.Read_Unread_List->query_read: "
			  "Internal error.\n", backtrace() }) );
	    i = k;
	    j++;
	  }
      }
#ifdef CHECK_LISTS
    check_state();
#endif
    return texts_in_conf;
  }

  /* Returnerar 1 om listan ndrades */
  int mark_as_read(int n)
  {
#ifdef CHECK_LISTS
    check_state();
#endif
    WERROR(sprintf("mark_as_read: n = %d, l::texts = %O\n", n, l::texts));
    if (!sizeof(l::texts))
      { /* Frsta lsta texten */
	l::texts = ({ n, n+1 });
	return 1;
      }
	    
    int i = binary_search(n, l::texts);
    if (i % 2)
      { /* Texten r inte markerad som lst sedan tidigare */
	WERROR(sprintf("i = %d\n", i));
	if (n == l::texts[i])
	  {
	    /* Frsta elementet i ett olsta-intervall */
	    l::texts[i]++;
	    if ( ((i+1) < sizeof(l::texts)) && (l::texts[i] == l::texts[i+1]) )
	      /* Enda elementet i intervallet frsvann */
	      l::texts = l::texts[..i-1] + l::texts[i+2..];
	    WERROR(sprintf("l::texts = %O\n", l::texts));
	  }
	else if ( ((i+1) < sizeof(l::texts))
		  && (l::texts[i + 1] == (n + 1)) )
	    
	  {
	    /* Sista elementet i ett lsta-intervall */
	    l::texts[i+1]--;
	  }
	else
	  {
	    l::texts = l::texts[..i] + ({ n, n + 1 }) + l::texts[(i + 1)..];
	  }
	return 1;
      }
#ifdef CHECK_LISTS
    check_state();
#endif
    return 0;
  }

  int next_unread_text(int n, array(int) texts_in_conf)
  {
#ifdef CHECK_LISTS
    check_state();
#endif
    int t;
    int i = binary_search(n, l::texts);
    if (i % 2)
      {
	/* Texten r olst */
	return n;
      }
    while ((++i < sizeof(l::texts)) && (t < sizeof(texts_in_conf)))
      {
	t = binary_search(l::texts[i], texts_in_conf);
	while (texts_in_conf[t] < l::texts[i])
	  {
	    /* Stega frbi lsta-intervallet */
	    t++;
	  }
	if ((i >= sizeof(l::texts) - 1)
	    || (texts_in_conf[t] < l::texts[i + 1]))
	  {
	    /* Texten finns i ett olst-intervall */
	    return texts_in_conf[t];
	  }
	else
	  {
	    /* G till nsta olsta-intervall */
	    i += 2;
	  }
      }
#ifdef CHECK_LISTS
    check_state();
#endif
    return 0;
  }
}

/* Information som alltid hlls i minnet */
class UPerson
{
  inherit DB_Object;
  inherit Types.Person;

  constant db_fields = ({ "username", "privileges",
			  "flags", "last_login",
			  "user_area", "total_time_present",
			  "sessions", "created_lines",
			  "created_bytes", "read_texts",
			  "no_of_text_fetches", "created_persons",
			  "created_confs", "first_created_local_no",
			  "no_of_created_texts", "no_of_marks",
			  "no_of_confs", "passwd" });
  string passwd;
}

/* Information om medlemskap. Behvs fr att hmta olsta, samt visa
 * status. */
class Person_memberships
{
  constant db_version = 2;
  array(object(Membership)) memberships = ({ });
  
  array to_db()
  {
    return Array.map(memberships, "to_db");
  }
  
  object from_db(array data)
  {
    int size = sizeof(data);
    memberships = allocate(size);
    for(int i = 0; i<size; i++)
      {
	if (!(memberships[i] = Membership()->from_db(data[i])))
	  return 0;
      }
    return this_object();
  }

  object from_db_version(int version, array data)
  {
    WERROR(sprintf("Server.Person_memberships->from_db_version %d.\n",
		   version));
    int size = sizeof(data);
    memberships = allocate(size);
    for(int i = 0; i<size; i++)
      {
	if (!(memberships[i] = Membership()
	      ->from_db_version(version, data[i])))
	  return 0;
      }
    return this_object();
  }
}

/* Skapade texter. Behvs mer sllan. Cachas. */
class Person_texts
{
  inherit DB_Object;

  inherit Text_List;
}

/* Olsta-info. Behvs enbart nr personen r inloggad. */
class Person_unread
{
  inherit Read_Unread_List;
}

class Mark
{
  inherit Types.Mark;
  inherit DB_Object;
  constant fields = ({ "text_no", "type" });
}

class Mark_List
{
  inherit Types.Mark_List;
  constant fields = ({ FIELD3("array", "contents", Mark) });

  array to_db()
  {
    return Array.map(contents, "to_db");
  }

  object from_db(array a)
  {
    int size = sizeof(a);
    contents = allocate(size);
    for (int i = 0; i < size; i++)
      if (!(contents[i] = Mark()->from_db(a[i])))
	return 0;
    return this_object();
  }
}

/* Behvs enbart fr att terse markeringar, och enbart nr personen r
 * inloggad. */
class Person_marks
{
  inherit Mark_List;
}

class Text
{
  inherit DB_Object;
  constant db_fields = ({ "text" });

  string text; /* Utan renderad */

  void create(string s)
  {
    text = s;
  }
}

class Misc_Info
{
  inherit Types.Misc_Info;
  inherit DB_Object;
  constant db_fields = ({ "type", "value" });

  void create(int t, int v)
  {
    type = t;
    value = v;
  }
}

class Text_Stat
{
  // inherit Types.Text_Stat;
  constant fields = ({ FIELD("time", "creation_time"),
		       FIELD("int", "author"),
		       FIELD("int", "no_of_lines"),
		       FIELD("int", "no_of_chars"),
		       FIELD("int", "no_of_marks") });
  
  inherit DB_Object;
  constant db_fields = ({ "creation_time", "author",
			  "no_of_lines", "no_of_chars",
			  "no_of_marks", "recipients",
			  "cc_recipients", "bcc_recipients",
			  "comment_tos", "commented_ins",
			  "footnote_tos", "footnoted_ins",
			  "subject", "no_of_system_marks" });
  
  time_t creation_time;
  int author;
  int no_of_lines;
  int no_of_chars;
  int no_of_marks;

  /* ({ huvud-misc-info, ({ relaterade misc-infon }) }) */
  array recipients = ({ });
  array cc_recipients = ({ });
  array bcc_recipients = ({ });
  array comment_tos = ({ });
  array commented_ins = ({ });
  array footnote_tos = ({ });
  array footnoted_ins = ({ });
  
  string subject;
  int no_of_system_marks; /* Interna markeringar av user-areor och
			   * presentationer. */

  array query_misc_info_type(array a, int type, int|void text_no)
  {
    array res = ({ });
    foreach(a, array m)
      {
	res += ({ Misc_Info(type, m[0]) });
	if (text_no)
	  /* Lgg till "lokalt" nummer. */
	  res += ({ Misc_Info(Types.misc_local_no, text_no) });

	res += Array.map(m[1..], Types.Misc_Info);
      }
    return res;
  }

  array query_misc_info(int text_no)
  {
    return query_misc_info_type(recipients, Types.misc_recipient, text_no)
      +  query_misc_info_type(cc_recipients, Types.misc_cc_recipient, text_no)
      +  query_misc_info_type(comment_tos, Types.misc_comment_to)
      +  query_misc_info_type(commented_ins, Types.misc_commented_in)
      +  query_misc_info_type(footnote_tos, Types.misc_footnote_to)
      +  query_misc_info_type(footnoted_ins, Types.misc_footnoted_in);
  }
  
#if 0
  array fancy_to_db()
  {
    return ({ Array.map(misc_info, "to_db"),
		({ }) /* Inga aux-info */ });
  }

  object fancy_from_db(array data)
  {
    if (sizeof(data) != 2)
      return 0;
    array a = data[0];
    int size = sizeof(a);
    misc_info = allocate(size);
    for (int i = 0; i<size; i++)
      if (!(misc_info[i] = Misc_Info()->from_db(a[i])))
	return 0;
    return this_object();
  }
#endif
}

class Session
{
  inherit Types.Static_Session_Info : stat;
  inherit Types.Dynamic_Session_Info : dynamic;

  time_t last_seen;
  
  array query_fields()
  {
    throw( ({ "Server.Session->query_fields: Should not be called!\n",
	      backtrace() }) );
  }

  int query_and_update_idle_time()
  {
    idle_time = (flags & Types.session_user_active_used)
                && (time() - last_seen);
    if (idle_time < 0)
      werror("Session.query_and_update_idle_time: idle_time < 0!\n");
    return idle_time;
  }
  
  string query_dynamic()
  {
    WERROR("Server.Session.query_dynamic\n");
    return query_protA(dynamic::query_fields());
  }

  string query_static()
  {
    return query_protA(stat::query_fields());
  }
}
